#RequireAdmin
#Region ;**** Directives created by AutoIt3Wrapper_GUI ****
#AutoIt3Wrapper_Icon=Joystick.ico
#AutoIt3Wrapper_Outfile=Tur-Game_Controller_Order.exe
#AutoIt3Wrapper_Res_Description=Tur-Game Controller Order
#AutoIt3Wrapper_Res_Fileversion=1.6.0.0
#AutoIt3Wrapper_Res_ProductVersion=1.6.0
#EndRegion ;**** Directives created by AutoIt3Wrapper_GUI ****

#include <Constants.au3>
#include <Array.au3>
#include <Misc.au3>

Local $Version = "1.6"

If _Singleton("Tur-Game_Controller_Order", 1) = 0 Then
	MsgBox(16, "Warning", "Tur-Game_Controller_Order.exe is already running - Exiting.", 5)
	Exit
EndIf

Global $HIDArrayRaw[0]
Global $HIDsFound = "0"

; Find out which devcon to run depending on OS architecture.
If @OSArch = "X64" Then
	Global $DevCon = @ScriptDir & "\devcon64.exe"
ElseIf @OSArch = "X86" Then
	Global $DevCon = @ScriptDir & "\devcon32.exe"
Else
	MsgBox(16, "Error", "Could not determine if this is an x64 or x86 machine.. ?!")
	Exit
EndIf

; Define the INI file.

If $CmdLine[0] > "0" Then
	Global $IniFile = @ScriptDir & "\" & $CmdLine[1]
	Local $IniName = $CmdLine[1]
	TraySetToolTip("Tur-Game Controller Order " & $Version & " (" & $CmdLine[1] & ")")
Else
	Global $IniFile = @ScriptDir & "\Settings.ini"
	Local $IniName = "Settings.ini"
	TraySetToolTip("Tur-Game Controller Order " & $Version & " (Settings.ini)")
EndIf

; Make sure we really have a devcon command...
If Not FileExists($DevCon) Then
	MsgBox(16, "Error", "Could not find " & $DevCon)
	Exit
EndIf

; Check if the ini file exists. If not, run the guide.
If Not FileExists($IniFile) Then

	; Run the Game Controllers control panel item.
	Run("control.exe joy.cpl")

	; Find all HIDs for everything called "game controller".
	Local $CurrentPID = Run(@ComSpec & ' /c ' & Chr(34) & $DevCon & Chr(34) & " status *game controller* | findstr /c:HID\VID /c:Name:", "", @SW_HIDE, $STDERR_CHILD + $STDOUT_CHILD)
	If Not ProcessWaitClose($CurrentPID, 60) Then
		MsgBox(16, "No output", "Did not get any return from " & $DevCon & " status *game controller*")
		Exit
	EndIf

	; Read the output we got.
	$CurrentOutput = StdoutRead($CurrentPID)
	If $CurrentOutput == "" Then
		MsgBox(16, "Error", "Did not find any game controllers when running" & @CRLF & Chr(34) & $DevCon & Chr(34) & " status *game controller*" & @CRLF & @CRLF & "Make sure your BlissBox and any other controllers you have are connected.")
		Exit
	EndIf

	; Show initial message with information.
	MsgBox(48, "Setup " & $IniName & " (version " & $Version & ")", "This is the first time you run this program. We will now run through the setup and identify your joystick ports HIDs." & @CRLF & @CRLF & "I have opened the Game Controllers window and you should see all your controllers. Please connect all Game Controllers and make sure they appear before clicking OK.")


	; Create a new array from the output from devcon, clean up of any newlines and empty rows that might appear.
	Local $HIDArrayRaw[0][2]
	Local $HIDArrayRawRaw = StringSplit($CurrentOutput, @CR)

	For $i = 1 To $HIDArrayRawRaw[0]
		Local $HIDCleaned = StringStripCR(StringStripWS($HIDArrayRawRaw[$i], 1))
		If StringInStr($HIDCleaned, "HID\VID", 0, 1, 1) Then
			Local $GotHID = $HIDCleaned
		EndIf
		If StringInStr($HIDCleaned, "Name:", 0, 1, 1) Then
			ConsoleWrite($HIDCleaned & " : " & $GotHID & @CRLF)
			If $HIDCleaned <> "" Then
				_ArrayAdd($HIDArrayRaw, $GotHID & "|" & $HIDCleaned)
			EndIf
		EndIf

	Next
	;_arraydisplay($HIDArrayRaw)
	Local $HIDsFound = UBound($HIDArrayRaw)

	; Verify that all controllers can be removed and readded.
	MsgBox(48, "Setup " & $IniName & " (version " & $Version & ")", "First we have to verify that all your game controllers are happy with being disconnected and reconnected with the DevCon command." & @CRLF & @CRLF & "I'll remove all Game Controllers and then readd them back. Click OK and stand by.")
	For $i = 0 To $HIDsFound - 1
		Local $HidNameTemp = StringSplit($HIDArrayRaw[$i][0], "&")
		Local $HidNameTemp2 = $HidNameTemp[1] & "&" & $HidNameTemp[2]
		Local $HidNameTemp3 = StringSplit($HidNameTemp2, "\") ; Part 1, get the part we need to search for the name in the registry.
		Local $HidNameTemp4 = $HidNameTemp3[2] ; Part 2
		ConsoleWrite("Removing HID: " & $HidNameTemp4 & @CRLF)
		RunWait(Chr(34) & $DevCon & Chr(34) & " remove " & Chr(34) & "*" & $HidNameTemp4 & "*" & Chr(34), "", @SW_HIDE) ; Actually discconect it.
	Next
	MsgBox(48, "Setup " & $IniName & " (version " & $Version & ")", "All controllers should now be gone. Click OK to rescan hardware (Might take a few seconds).")
	ConsoleWrite("Rescanning hardware..." & @CRLF)
	RunWait(Chr(34) & $DevCon & Chr(34) & " rescan", "", @SW_HIDE)						; Rescan the hardware so all ports pop back into the Game Controller window.
	SplashTextOn("Device installations", "Waiting for possible device installations." & @CRLF & "Stand by...", 300, 65)
	Sleep(5000)
	SplashTextOn("Device installations", "Waiting for device installations." & @CRLF & "Stand by...", 300, 65)
	WinWaitClose("Device Setup", "", 45)
	SplashOff()
	Local $Answer = MsgBox(36, "Setup " & $IniName & " (version " & $Version & ")", "All controllers should now be back. If you are missing any controllers, you can not use this program with them connected. Did all controllers come back OK?")
	ConsoleWrite($Answer & @CRLF)
	If $Answer <> 6 Then
		MsgBox(16, "Ugh", "Sorry to hear that. Please reconnect all lost controllers now to get them back and throw this program in the trash." & @CRLF & @CRLF & "If you find a DevCon command to get them back, let me know...")
		Exit
	EndIf

	; Create a nice list we can use in the next msgbox.
	Local $HIDDisplay = ""
	Local $HIDDisplayArray[0]
	Local $HIDCleaned[0][2]

	For $i = 0 To $HIDsFound - 1

		; Make sure the HID contains atleast 2 backslashes or its nothing we want.. weird
		If Not StringInStr($HIDArrayRaw[$i][0], "\", 0, 2) Then
			MsgBox(16, "HID Warning", "The HID:" & @CRLF & $HIDArrayRaw[$i][0] & @CRLF & "does not contain atleast 2 backslashes. Will try anyway but something is strange...")
		EndIf

		Local $HidNameTemp = StringSplit($HIDArrayRaw[$i][0], "&")
		Local $HidNameTemp2 = $HidNameTemp[1] & "&" & $HidNameTemp[2]
		Local $HidNameTemp3 = StringSplit($HidNameTemp2, "\") ; Part 1, get the part we need to search for the name in the registry.
		Local $HidNameTemp4 = $HidNameTemp3[2] ; Part 2
		Local $HIDNamed = RegRead("HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\MediaProperties\PrivateProperties\Joystick\OEM\" & $HidNameTemp4, "OEMName")
		If $HIDNamed == "" Then
			$HIDNamed = StringStripWS(StringReplace($HIDArrayRaw[$i][1], "Name:", ""),1)
			MsgBox(64, "Using Alternate name..", "Did not find the name for " & @CRLF & $HIDArrayRaw[$i][0] & @CRLF & @CRLF & "However, devcon calls it " & Chr(34) & $HIDNamed & Chr(34) & " so that is the name I will present it as (Might not match the Game Controllers window name).")
		EndIf
		ClipPut($HIDArrayRaw[$i][0])
		Local $SelectedHIDToUse = InputBox($HIDNamed, "Select the unique part of the HID to use for this device." & @CRLF & "The default is in most cases enough." & @CRLF & @CRLF & "In your clipboard is the full HID so just paste it and copy the unique part. Full HID in clipboard: " & @CRLF & $HIDArrayRaw[$i][0] & @CRLF & @CRLF & "To test if it is valid, start CMD as Administrator and go to the location of the devcon command. Run:" & @CRLF & "devcon remove " & Chr(34) & "*" & $HidNameTemp4 & "*" & Chr(34) & @CRLF & "That controller (only that controller) should vanish." & @CRLF & @CRLF & "Run:" & @CRLF & "devcon rescan" & @CRLF & "and it should pop back", $HidNameTemp4,"",400,310)
		If $SelectedHIDToUse == "" Then
			Exit
		EndIf
		_ArrayAdd($HIDDisplayArray, $HIDNamed & @TAB & "HID: " & $SelectedHIDToUse)
		_ArrayAdd($HIDCleaned, $SelectedHIDToUse & "|" & $HIDNamed)
	Next
	_ArraySort($HIDDisplayArray) ; Sort the list before we present it. Looks better.
	Local $HIDDisplay = _ArrayToString($HIDDisplayArray, @CR) ; Convert the sorted array to a string that we can display.

	; Show the next message, that we are ready to start.
	Local $GoResponse = MsgBox(65, "Ready!", "The following controllers were found:" & @CRLF & @CRLF & $HIDDisplay & @CRLF & @CRLF & "Press OK to continue.")
	; Make sure OK was clicked. Quit otherwise.
	If $GoResponse <> "1" Then
		Exit
	EndIf

	Local $DonePorts = "|"
	Local $DonePortsNice = ""

	_ArraySort($HIDCleaned)

	Local $HIDsFound = UBound($HIDCleaned)


	; Go through each HID that we found.
	For $i = 0 To $HIDsFound - 1
		; Put the HID in a string (easier to work with).
		Local $HID = $HIDCleaned[$i][0]
		Local $HIDn = $HIDCleaned[$i][1]
		; Make sure its not empty. It should never be, but here is a check for it since we can.QZ<
		If $HID <> "" Then

			; Ask which priority this game controller should be in the future. 1 and up.
			Local $SelectedPort = InputBox($HID, "Controller found: " & Chr(34) & $HIDn & Chr(34) & @CRLF & @CRLF & "Where in the Game Controllers list would you like this to appear in the future?" & @CRLF & "1 for the top (first), " & $HIDsFound & " for the bottom." & @CRLF & @CRLF & "Or in other words, which Joy # would you like this controller to appear as?", "", "", 450)

			While StringIsDigit($SelectedPort) = "0" Or $SelectedPort < "1" Or $SelectedPort > $HIDsFound Or StringInStr($DonePorts, "|" & $SelectedPort & "|") And $SelectedPort <> ""
				If StringInStr($DonePorts, "|" & $SelectedPort & "|") Then
					MsgBox(16, "#" & $SelectedPort & " already assigned!", "Already selected a controller #" & $SelectedPort & @CRLF & @CRLF & $DonePortsNice & @CRLF & "Please try again")
				EndIf
				Local $SelectedPort = InputBox($HIDn, "Wrong input. Should be a number from 1 to " & $HIDsFound & @CRLF & "Please select another port for " & $HIDn & @CRLF & @CRLF & "1 for the top (first), " & $HIDsFound & " for the bottom." & @CRLF & @CRLF & "Or in other words, which Joy # would you like this controller to appear as?", "", "", 450)
			WEnd

			$DonePorts = $DonePorts & $SelectedPort & "|"
			$DonePortsNice = $DonePortsNice & $SelectedPort & " = " & $HIDn & @CRLF

			; Depending on what the user responded with, write it to the Settings.ini file.
			If $SelectedPort == "" Then
				MsgBox(16, "Aborting.", "Nothing typed. Aborting.")
				If FileExists($IniFile) Then
					FileDelete($IniFile)
				EndIf
				Exit
			Else
				IniWrite($IniFile, "Game Ports", "Port" & $SelectedPort, $HID)
			EndIf
		EndIf
	Next

	; Sort the INI file. Looks better.
	_IniSort($IniFile)

	; Add extra options to ini
	FileWriteLine($IniFile, "")
	FileWriteLine($IniFile, "")
	IniWrite($IniFile, "Daemon Settings", "DaemonMode", "FALSE")
	IniWrite($IniFile, "Daemon Settings", "DaemonModeDelay", "10")

	; Show a final message and exit.
	MsgBox(0, "Done", "Setup is done. Providing all went well, you should now be able to restart me to have the ports reordered automatically." & @CRLF & @CRLF & "Please verify that your ports in " & $IniName & " are correct, starting with Port1 and going up without gaps." & @CRLF & @CRLF & "To restart this setup, delete the " & $IniName & " file" & @CRLF & @CRLF & $DonePortsNice)

	Exit

EndIf

; We end up here if you start the program with a Settings.ini file existing already.

; Read the Ports from the ini file into an array.
Global $PortArray = ""
Global $PortArray = IniReadSection($IniFile, "Game Ports")

; Read other stuff from INI
Global $DaemonMode = StringUpper(IniRead($IniFile, "Daemon Settings", "DaemonMode", "FALSE"))
Global $DaemonModeDelay = IniRead($IniFile, "Daemon Settings", "DaemonModeDelay", "10")

; Verify that we got an array
If Not IsArray($PortArray) Then
	MsgBox(16, "Error", "Settings.ini error. Can not read ports. Please delete your ini file and re to set it up correctly.")
	Exit
ElseIf $PortArray[0][0] < "2" Then
	MsgBox(16, "Error", "Error in Settings.ini. Seems there are not atleast 2 ports defined. Please delete the .ini and rerun to set it up correctly.")
	Exit
EndIf

If $DaemonMode = "FALSE" Then
	Local $Returns = _FindConnectedHIDS()
	If $Returns = "NONE_FOUND" Then
		MsgBox(16, "Error", "Did not find any game controllers when running" & @CRLF & Chr(34) & $DevCon & Chr(34) & " status *game controller*" & @CRLF & @CRLF & "Make sure your controllers are connected.", 10)
		Exit
	EndIf
	_ReorderControllers()
	_CleanupNonDefinedPorts()

	If StringInStr($Version, "b") <> 0 Then
		ConsoleWrite("Beta version pausing for 10 seconds..." & @CRLF)
		Sleep(10000)
	EndIf

Else
	; Do one initial fix at startup.
	Local $Returns = _FindConnectedHIDS()
	If $Returns <> "NONE_FOUND" Then
		_ReorderControllers()
		_CleanupNonDefinedPorts()
	EndIf
	Sleep($DaemonModeDelay * 1000)

	; Start looping.
	While 1

		While WinExists("Device Setup")
			ConsoleWrite("Waiting for Device Setup window to go away." & @CRLF)
			Sleep(1000)
		WEnd

		Global $OldHIDArrayRaw = $HIDArrayRaw
		Global $OldHIDsFound = $HIDsFound

		Local $Returns = _FindConnectedHIDS()
		If $Returns <> "NONE_FOUND" Then
			ConsoleWrite("We found " & $HIDsFound & " controllers! Lets check if they are the same as last time")
			If _CompareArray($HIDArrayRaw, $OldHIDArrayRaw) = 1 Then
				ConsoleWrite(" - its not. Reorder time!" & @CRLF)
				_ReorderControllers()
				_CleanupNonDefinedPorts()
			Else
				ConsoleWrite(" - Same as before. Do nothing." & @CRLF)
			EndIf
		Else
			ConsoleWrite("No controllers found. Skipping this turn." & @CRLF)

		EndIf
		Sleep($DaemonModeDelay * 1000)
	WEnd


EndIf


; Verify that the controllers are connected
Func _FindConnectedHIDS()
	; Find all HIDs for everything called "game controller".
	Local $CurrentPID = Run(@ComSpec & ' /c ' & Chr(34) & $DevCon & Chr(34) & " status *game controller* | findstr HID\VID", "", @SW_HIDE, $STDERR_CHILD + $STDOUT_CHILD)
	If Not ProcessWaitClose($CurrentPID, 60) Then
		MsgBox(16, "No output", "Did not get any return from " & $DevCon & " status *game controller*", 10)
		Exit
	EndIf
	; Read the output we got.
	$CurrentOutput = StdoutRead($CurrentPID)
	If $CurrentOutput == "" Then
		Global $HIDArrayRaw[0]
		Global $HIDsFound = "0"
		Return ("NONE_FOUND")
	EndIf

	; Create a new array from the output from devcon so we know which HIDs actually exists.
	Global $HIDArrayRaw[0]
	Local $HIDArrayRawRaw = StringSplit($CurrentOutput, @CR)
	For $i = 1 To $HIDArrayRawRaw[0]
		Local $HIDCleaned = StringStripWS($HIDArrayRawRaw[$i], 1)
		If $HIDCleaned <> "" Then
			Local $HIDArrayTemp = StringSplit($HIDCleaned, "\")
			Local $HIDCleaned = $HIDArrayTemp[2]
			If $HIDCleaned <> "" Then
				_ArrayAdd($HIDArrayRaw, $HIDCleaned)
			EndIf
		EndIf
	Next
	Global $HIDsFound = UBound($HIDArrayRaw)
	Return ("FOUND")
EndFunc   ;==>_FindConnectedHIDS

Func _ReorderControllers()
	Local $HIDsDone[0]
	Local $PortNum = "1"
	For $y = 1 To $PortArray[0][0]																; We need to go through this the number of ports in Settings.ini times.. *
		Local $ActuallyRemovedSomething = "FALSE"												; If this is still FALSE further down, there will be no need to rescan the hardware.
		For $i = 1 To $PortArray[0][0]															; * the time the number of ports. Confusing, I know.
			Local $Name = $PortArray[$i][0]														; Get the defined name from Settings.ini
			Local $ID = $PortArray[$i][1]														; Get the defined HID from Settings.ini

			If _ArraySearch($HIDArrayRaw, $ID) <> "-1" Then										; Check if this HID is actually detected right now before trying to disconnect it.
				If $Name <> "Port" & $PortNum And _ArraySearch($HIDsDone, $Name) = "-1" Then	; Dont disconnect this one if its the current port or if its already been disconnected before.
					ConsoleWrite("Removing " & $Name & " which has ID: " & $ID & @CRLF)
					RunWait(Chr(34) & $DevCon & Chr(34) & " remove " & Chr(34) & "*" & $ID & "*" & Chr(34), "", @SW_HIDE) ; Actually discconect it.
					$ActuallyRemovedSomething = "TRUE" 											; We actually removed something. Set this to TRUE so we know we should rescan the hardware.
				Else
					_ArrayAdd($HIDsDone, $Name) 												; This port is the current port and its now done. Add it to array so its not run again.
				EndIf
			EndIf
		Next
		$PortNum = $PortNum + 1																	; Add +1 to the Portnumber so we check the next one in order on the next loop.
		If $ActuallyRemovedSomething = "TRUE" Then												; If this is TRUE, we actually removed something up above and need to rescan.
			ConsoleWrite("- Rescanning hardware" & @CRLF)
			RunWait(Chr(34) & $DevCon & Chr(34) & " rescan", "", @SW_HIDE)						; Rescan the hardware so all ports pop back into the Game Controller window.
		EndIf
	Next
EndFunc   ;==>_ReorderControllers

Func _CleanupNonDefinedPorts()
	; Clean up any other controllers that are not defined in the ini file and put them at the bottom of the Game Controllers list..
	Local $FoundExtras = "FALSE"																						; Set this to FALSE so we know if we need to rescan hardware further down.
	For $i = 0 To $HIDsFound - 1																						; For every HID found connected to the system...
		If _ArraySearch($PortArray, $HIDArrayRaw[$i]) = -1 Then															; Check if its defined in the Settings.ini file currently used.
			Local $HIDNamed = RegRead("HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\MediaProperties\PrivateProperties\Joystick\OEM\" & $HIDArrayRaw[$i], "OEMName")
			If $HIDNamed == "" Then
				$HIDNamed = "Unknown"
			EndIf
			ConsoleWrite("Found unspecified controller: " & $HIDNamed & " - Disconnecting it so its put last: " & $HIDArrayRaw[$i] & @CRLF)
			RunWait(Chr(34) & $DevCon & Chr(34) & " remove " & Chr(34) & "*" & $HIDArrayRaw[$i] & "*" & Chr(34), "", @SW_HIDE) ; Actually discconect it as it was not in the Settings.ini file.
			Local $FoundExtras = "TRUE"																					; Set this so we know we actually disconnected something and need to rescan hardware.
		EndIf
	Next
	If $FoundExtras = "TRUE" Then																						; Check if we actually disconnected something.
		ConsoleWrite("- Rescanning hardware" & @CRLF)
		RunWait(Chr(34) & $DevCon & Chr(34) & " rescan", "", @SW_HIDE)													; Rescan the hardware so all ports pop back into the Game Controller window.
	EndIf
EndFunc   ;==>_CleanupNonDefinedPorts

; Function to sort the ini file at the end of creation. Not needed but looks better in Settings.ini if they are in order from 1-4.
Func _IniSort($hIni)
	Local $aIRSN = IniReadSectionNames($hIni)
	If Not IsArray($aIRSN) Then Return SetError(1, 0, 0)
	_ArraySort($aIRSN, 0, 1)
	Local $aKey, $sHold
	For $iCC = 1 To UBound($aIRSN) - 1
		Local $aIRS = IniReadSection($hIni, $aIRSN[$iCC])
		If Not IsArray($aIRS) Then ContinueLoop
		For $xCC = 1 To $aIRS[0][0]
			$aKey &= $aIRS[$xCC][0] & Chr(1)
		Next
		If $aKey Then
			$aKey = StringSplit(StringTrimRight($aKey, 1), Chr(1))
			_ArraySort($aKey, 0, 1)
			$sHold &= '[' & $aIRSN[$iCC] & ']' & @CRLF
			For $aCC = 1 To UBound($aKey) - 1
				$sHold &= $aKey[$aCC] & '=' & IniRead($hIni, $aIRSN[$iCC], $aKey[$aCC], 'blahblah') & @CRLF
			Next
			$aKey = ''
		EndIf
	Next
	If $sHold Then
		$sHold = StringTrimRight($sHold, 2)
		FileClose(FileOpen($hIni, 2))
		FileWrite($hIni, $sHold)
		Return 1
	EndIf
	Return SetError(1, 0, 0)
EndFunc   ;==>_IniSort

; Function to compare two simple 1D arrays with each other.
Func _CompareArray($a, $b)
	If NOT IsArray($a) or NOT IsArray($b) Then
		Return 1
	EndIf
	If UBound($a) <> UBound($b) Then
		Return 1
	EndIf

	For $x = 0 To UBound($a) - 1
		If _ArraySearch($b, $a[$x]) = -1 Then
			Return 1
		EndIf
	Next

	For $x = 0 To UBound($b) - 1
		If _ArraySearch($a, $b[$x]) = -1 Then
			Return 1
		EndIf
	Next


;	For $x = 0 To UBound($a) - 1
;		If $a[$x] <> $b[$x] Then
;			Return 1
;		EndIf
;	Next


	Return 0
EndFunc   ;==>_CompareArray
